@@ -117,11 +117,13 @@ struct write_window {
 };
 
 static pthread_mutex_t window_lock = PTHREAD_MUTEX_INITIALIZER;
+static pthread_mutex_t window_fd_lock = PTHREAD_MUTEX_INITIALIZER;
 static struct write_window window[NR_WINDOWS];
 
 static int
 evict_writes (int fd, uint64_t offset, size_t len)
 {
+  ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&window_fd_lock);
   struct write_window oldest = { 0 };
 
   {
@@ -147,10 +149,14 @@ evict_writes (int fd, uint64_t offset, size_t len)
     window[NR_WINDOWS-1].offset = offset;
     window[NR_WINDOWS-1].len = len;
   }
-  /* Release lock here. */
+  /* Release window_lock here, but hold window_fd_lock since we don't
+   * want file_close -> remove_fd_from_window to close oldest.fd
+   */
 
   /* Evict the oldest window from the page cache (synchronously). */
   if (oldest.len > 0) {
+    int r;
+
     if (sync_file_range (oldest.fd, oldest.offset, oldest.len,
                          SYNC_FILE_RANGE_WAIT_BEFORE |
                          SYNC_FILE_RANGE_WRITE |
@@ -158,9 +164,12 @@ evict_writes (int fd, uint64_t offset, size_t len)
       nbdkit_error ("sync_file_range: cache=none: evicting oldest window: %m");
       return -1;
     }
-    if (posix_fadvise (oldest.fd, oldest.offset, oldest.len,
-                       POSIX_FADV_DONTNEED) == -1)
+    r = posix_fadvise (oldest.fd, oldest.offset, oldest.len,
+                       POSIX_FADV_DONTNEED);
+    if (r != 0) {
+      errno = r;
       nbdkit_debug ("posix_fadvise: POSIX_FADV_DONTNEED: (ignored): %m");
+    }
   }
 
   return 0;
@@ -172,7 +181,7 @@ evict_writes (int fd, uint64_t offset, size_t len)
 static void
 remove_fd_from_window (int fd)
 {
-  ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&window_lock);
+  ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&window_fd_lock);
   size_t i;
 
   for (i = 0; i < NR_WINDOWS; ++i)
@@ -200,10 +209,7 @@ file_unload (void)
   free (directory);
 }
 
-/* Called for each key=value passed on the command line.  This plugin
- * only accepts file=<filename> and dir=<dirname>, where exactly
- * one is required.
- */
+/* Called for each key=value passed on the command line. */
 static int
 file_config (const char *key, const char *value)
 {
@@ -321,7 +327,7 @@ file_config_complete (void)
   switch (mode) {
   case mode_none:
     nbdkit_error ("you must supply [file=]<FILENAME>, "
-                  "dir=<DIRNAME> or fd=<FD> "
+                  "dir=<DIRNAME>, fd=<FD> or dirfd=<FD> "
                   "parameter after the plugin name "
                   "on the command line");
     return -1;
@@ -688,8 +694,10 @@ file_open (int readonly)
     /* This is a hint so we ignore failures. */
 #ifdef HAVE_POSIX_FADVISE
     int r = posix_fadvise (h->fd, 0, 0, fadvise_mode);
-    if (r == -1)
-      nbdkit_debug ("posix_fadvise: %s: %m (ignored)", h->name);
+    if (r != 0) {
+      errno = r;
+      nbdkit_debug ("posix_fadvise: %s: (ignored): %m", h->name);
+    }
 #else
     nbdkit_debug ("fadvise is not supported");
 #endif
@@ -762,6 +770,21 @@ file_open (int readonly)
   h->can_blkzeroout = h->is_block_device;
   h->can_blkdiscard = h->is_block_device;
 
+  nbdkit_debug ("%s: fd=%d mode=%o "
+                "%s%s%s%s%s%s%s"
+                "sector_size=%d rotational=%s "
+                "blocksizes=(%" PRIu32 ",%" PRIu32 ",%" PRIu32 ")",
+                h->name, h->fd, h->statbuf.st_mode,
+                h->is_block_device ? "block_device " : "",
+                h->can_write       ? "can_write " : "readonly ",
+                h->can_punch_hole  ? "can_punch_hole " : "",
+                h->can_zero_range  ? "can_zero_range " : "",
+                h->can_fallocate   ? "can_fallocate " : "",
+                h->can_blkdiscard  ? "can_blkdiscard " : "",
+                h->can_blkzeroout  ? "can_blkzeroout " : "",
+                h->sector_size, h->rotational ? "true" : "false",
+                h->minimum, h->preferred, h->maximum);
+
   return h;
 
  err:
@@ -1029,6 +1052,7 @@ file_zero (void *handle, uint32_t count, uint64_t offset, uint32_t flags)
       return -1;
     }
 
+    ONCE (nbdkit_debug ("%s: setting %s=false", h->name, "can_punch_hole"));
     h->can_punch_hole = false;
   }
 #endif
@@ -1066,6 +1090,7 @@ file_zero (void *handle, uint32_t count, uint64_t offset, uint32_t flags)
         return -1;
       }
 
+      ONCE (nbdkit_debug ("%s: setting %s=false", h->name, "can_blkdiscard"));
       h->can_blkdiscard = false;
     }
     else {
@@ -1075,6 +1100,7 @@ file_zero (void *handle, uint32_t count, uint64_t offset, uint32_t flags)
         return -1;
       }
 
+      ONCE (nbdkit_debug ("%s: setting %s=false", h->name, "can_fallocate"));
       h->can_fallocate = false;
     }
   }
@@ -1098,6 +1124,7 @@ file_zero (void *handle, uint32_t count, uint64_t offset, uint32_t flags)
       return -1;
     }
 
+    ONCE (nbdkit_debug ("%s: setting %s=false", h->name, "can_zero_range"));
     h->can_zero_range = false;
   }
 #endif
@@ -1127,6 +1154,7 @@ file_zero (void *handle, uint32_t count, uint64_t offset, uint32_t flags)
         return -1;
       }
 
+      ONCE (nbdkit_debug ("%s: setting %s=false", h->name, "can_fallocate"));
       h->can_fallocate = false;
     } else {
       if (!is_enotsup (errno)) {
@@ -1135,6 +1163,7 @@ file_zero (void *handle, uint32_t count, uint64_t offset, uint32_t flags)
         return -1;
       }
 
+      ONCE (nbdkit_debug ("%s: setting %s=false", h->name, "can_punch_hole"));
       h->can_punch_hole = false;
     }
   }
@@ -1160,6 +1189,7 @@ file_zero (void *handle, uint32_t count, uint64_t offset, uint32_t flags)
       return -1;
     }
 
+    ONCE (nbdkit_debug ("%s: setting %s=false", h->name, "can_blkzeroout"));
     h->can_blkzeroout = false;
   }
 #endif
@@ -1200,6 +1230,7 @@ file_trim (void *handle, uint32_t count, uint64_t offset, uint32_t flags)
         return -1;
       }
 
+      ONCE (nbdkit_debug ("%s: setting %s=false", h->name, "can_punch_hole"));
       h->can_punch_hole = false;
     }
   }
@@ -1224,6 +1255,7 @@ file_trim (void *handle, uint32_t count, uint64_t offset, uint32_t flags)
         return -1;
       }
 
+      ONCE (nbdkit_debug ("%s: setting %s=false", h->name, "can_blkdiscard"));
       h->can_blkdiscard = false;
     }
   }
@@ -1259,7 +1291,9 @@ file_can_extents (void *handle)
   ACQUIRE_LOCK_FOR_CURRENT_SCOPE (&lseek_lock);
   r = lseek (h->fd, 0, SEEK_HOLE);
   if (r == -1) {
-    nbdkit_debug ("extents disabled: lseek: SEEK_HOLE: %m");
+    nbdkit_debug ("lseek: SEEK_HOLE: error (ignored): %m");
+    nbdkit_debug ("extents disabled because the kernel does not support "
+                  "lseek and SEEK_HOLE for this file type");
     return 0;
   }
   return 1;
