@@ -32,6 +32,10 @@
 
 /* Functions and constants in the virtual nbdkit.* module. */
 
+/* Python is broken, this MUST come first. */
+#define PY_SSIZE_T_CLEAN 1
+#include <Python.h>
+
 #include <config.h>
 
 #include <stdio.h>
